################## Plot Library
##
## This library contains plot objects [and their possible exeptions]
## A plot is an object which take a series as input and [connecting to plotfairy] produce an image of the given data
##
## This file is HTML-free.
## TO DO:
#####################################
import time,random,socket,urllib,json,cherrypy,calendar,base64   
from Utils import *
from SeriesLib import *
from types import *
import urllib2

class EmptyPlotException(Exception):
  """
   This exception is raised when there is no enough data to build a plot
  """
  pass

class PlotfairyException(Exception):
  """
  This exception is raised if there is any problem in the connection with the Plotfairy service
  """
  pass

class Plot(object):
  """
   Plot object represents a single plot (of different kinds).
  """
  plotType=''
  plotTemplate={}
  
  def __init__(self,config,inputSeries=None,legenda=None,):
    """
    To make a plot I need: ServieConfiguration object (to have plotfairy URL), inputseries to build the plot data and there may be a legend
    """
    self.inputSeries=inputSeries
    self.legenda=legenda
    self.loadingURL=""
    self.enoughData=False
    self.config=config

  def getPlotImg(self):
    """
    This method produce the json object for plotfairy and load its result as a base64 image
    """
    self.buildPlotData() ## special processing/upgrade of the local plotTemplate (overridden by the subclasses)
    if not(self.enoughData): 
       print "Error:: no enaugh data for the plot!"
       raise EmptyPlotException()
       return
    
    url = self.config.plotfairyUrl+"plot"
    # data needs to be a json object (since it could be big I need to pass it via POST)
    values = {'type': self.plotType,
              'data': json.dumps(self.plotTemplate) }
    data = urllib.urlencode(values)
    req = urllib2.Request(url,data)
    try:
      # for the lock needs see comment in T0Mon.py
      self.config.plotfairyLock.acquire()
      response = urllib2.urlopen(req)
    except (urllib2.HTTPError,urllib2.URLError), e:
      # if plotfairy is not working I might want to warn the user and the service administrator
      raise PlotfairyException()
    finally:
      self.config.plotfairyLock.release()
    image=response.read()
    # since I don't want any HTTP call for the images (they could go to another frontend) I need to push them in the html
    # to do so I need the base64 encoding
    data=base64.b64encode(image)
    return data
          
         
  def buildPlotData():
    """
     This is a 'virtual' method, must pe overridden in the subclasses in order to fill the PlotTemplate 
    """
    pass
     
      

############################################################################################################################### Scatter
class ScatterPlot(Plot):
   """
    Tier page plots. Input Series= DatasetSeries  
   """ 
   plotType='Scatter'
   marker="o"
   #plot template is filled during the building
   plotTemplate={ "draw_lines": False,  
                "legend": "bottomleft",
                "xaxis": {
                        "format": "time", 
                        "label": "time" }, 
                "yaxis": {
                        "format": "num", 
                        "label": "",
                        "log": True 
                         }}
                         
   def makeTitle(self,hours,interval):
     """
     special method reserved fot ScattePlots
     """
     self.title=self.title%(hours,interval)
                          
   def buildPlotData(self):
     try:
      self.series=[]
      for (name,d) in self.inputSeries.items():
          if ('log' in self.plotTemplate['yaxis']) and self.plotTemplate['yaxis']['log']:
             x,y= self.clearZeroValues(getattr(d,self.xAttr),getattr(d,self.yAttr))
          else:
            x=getattr(d,self.xAttr)
            y=getattr(d,self.yAttr)
            
          #if len(getattr(d,self.xAttr))!=len(getattr(d,self.yAttr)): 
            #print "error: different data length!!"+self.xAttr+"  "+self.yAttr
            
          if len(getattr(d,self.xAttr))>0:
            self.enoughData=True
            self.series.append({ 
                "colour": self.legenda.getColor(name), 
                "label": name, 
                "marker": self.marker, 
                "x": x, 
                "y": y })
      self.plotTemplate.update({"series": self.series,"title": self.title})
      self.plotTemplate['yaxis']['label']=self.yLabel
     except Exception, e:
       print e 

   def clearZeroValues(self,x,y):
     """
      This method is used to remove 0 values: if the plot use a Logaritimic scale, 0 values are problematic
     """
     # debugging check!
     if len(x)!=len(y): 
        print "error: different data length"
        return x,y
     xx=[]
     yy=[]
     for i in range(0,len(x)):
       if y[i]>0.0001:
         xx.append(x[i])
         yy.append(y[i])
     return xx,yy
      
##################################
class EventsPlot(ScatterPlot):
    title="Event Rate [%dh] - AVG on %d-min intervals"
    yLabel="Hz"
    xAttr="eventsIntervalX"
    yAttr="eventsIntervalY"

    
class EventSizesPlot(ScatterPlot):
    title="Event Sizes Over Time [%dh] - AVG on %d-min intervals"
    yLabel="KB"
    xAttr="eventSizesX"
    yAttr="eventSizesY"   

class DataRatePlot(ScatterPlot):
    title="Data Rate Over Time [%dh] - AVG on %d-min intervals"
    yLabel="MB/s"
    xAttr="dataRateX"
    yAttr="dataRateY"
    
    def makeTitle(self,hours,interval):
     #dataRatePlot has always a interval [and the smaller is 1 minute]
     self.title=self.title%(hours,max(interval,1))    
##########################################3
## events and eventSize plot WITHOUT INTERVAL [raw data]
class EventsPlotNoInt(ScatterPlot):
    """
     Special Version of EventsPlot with no intervals (all values are printed)
    """
    title="Event Rate [%dh]"
    yLabel="Hz"
    xAttr="X"
    yAttr="Y"
    def makeTitle(self,hours,step):
     self.title=self.title%(hours)
    
class EventSizesPlotNoInt(EventsPlotNoInt):
    """
    Special Version of EventSizes with no intervals (all values are printed)
    """
    title="Event Sizes Over Time [%dh]"
    yLabel="KB"
    xAttr="X"
    yAttr="allEventSizesY"   


################################################################################################### BAR   -- plots
class BarPlot(Plot):
  """
  Statistical page plot. Input Series= IntervalSeries
  """
  enoughData=True
  title=" Change me!"
  color="#ffffff"
  plotType='Bar'
  plotTemplate={
    #"legend": "bottomleft", 
    "series": [
        {
            "colour": "#46cf51",  
            "values": []
        }
    ], 
    "sort": "value", 
    "title": "UpdateMe", 
    "xaxis": {
        "bins": None, 
        "format": "time", 
        "label": "Time", 
        "min": None,
        "max": None, 
        "width": None
    }, 
    "yaxis": {
        "format": "dec", 
        "label": "AVG # of job", 
        "log": False
    }
}
  def buildPlotData(self):
     dataSize=len(self.inputSeries.getYValues())
     #check data
     for i in self.inputSeries.getYValues():
       if i>0:
         self.enoughData=True
         break

     if dataSize>0 and self.enoughData:
       #print "PLOT=="+str(self.inputSeries.getYValues())
       self.plotTemplate['series'][0]['values']=self.inputSeries.getYValues()
       # some values need to be set
       self.plotTemplate['series'][0]['colour']=self.color 
       self.plotTemplate['xaxis']['bins']=dataSize
       self.plotTemplate['xaxis']['min']=self.inputSeries.minValue
       self.plotTemplate['xaxis']['max']=self.inputSeries.maxValue
       self.plotTemplate['xaxis']['width']=self.inputSeries.interval
       self.plotTemplate['title']=self.title
       #self.plotTemplate['xaxis']['label']=self.xaxisLabel
     
     
class ExpressBarPlot(BarPlot):
  title="Express Jobs"
  color="#e0ffff"
  
class ExpressMergeBarPlot(BarPlot):
  title="Express Merge"
  color="#e0ffff"
  
class RepackBarPlot(BarPlot):  
  title="Repack Jobs"
  color="#FFEFD5"
  
class RepackMergeBarPlot(BarPlot):  
  title="Repack Merge Jobs"
  color="#FFEFD5"

class RecoBarPlot(BarPlot):  
  title="Reco Jobs"
  color="#F0FFF0"
  
class RecoMergeBarPlot(BarPlot):  
  title="Reco Merge Jobs"
  color="#F0FFF0"
  
class AlcaBarPlot(BarPlot):  
  title="Alca Jobs"
  color="#FFFFE0"
  
class AlcaMergeBarPlot(BarPlot):  
  title="Alca Merge Jobs"
  color="#FFFFE0"
  
class AllJobsBarPlot(BarPlot):  
  title="All Jobs"
  color="#00A000"

########################################################################################################    
class Legend(object):
  """
   This object is used to generate a set of colors which is then used for all the plots in the same raw (tier plot page)
  """
    
  def __init__(self,set):
    """
     this method choose for every element in set (dataset or stream) a random color
    """
    self.colors={}
    for d in set:
      self.colors.update({d:self.random_color()})
             
  def random_color(self):
    return '#%02x%02x%02x'%(random.randint(0,255),random.randint(0,255),random.randint(0,255))
     
  def getColor(self,name):
    """
     return the color for the input dataset/stream 
    """
    return self.colors[name]
    
  def getHTML(self):
    """
    print the legend in HTML format
    """
    res="<BR/><BR/><BR/><BR/><B>Legenda:</B><BR/><BR/>"
    for k,v in self.colors.items():
      res+="""<B><FONT COLOR="%s">%s</FONT></B><BR/>"""%(v,k)
    return res
      
############################################################################# 
   

